@@ -12,7 +12,9 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use crate::utils::{get_keyword_safe_ident, to_camel_case, upgrade};
+use crate::utils::{
+    convert_octal_escape_to_rust_style_escape, get_keyword_safe_ident, to_camel_case, upgrade,
+};
 use crate::wrappers::{self, FieldType};
 use crate::{ErrorKind, Result};
 use ::askama::Template;
@@ -61,27 +63,28 @@ impl MessagesAndEnums {
 
 struct Message {
     ident: String,
-    proto_name: String,
     trait_ident: String,
     trait_delegate_macro_ident: String,
     submodule_ident: String,
     nested: MessagesAndEnums,
     fields: Vec<Field>,
-    fields_len: usize,
     oneofs: Vec<Oneof>,
+    bitfield_len: i32,
     simple_ident: String,
     single_field_ident: String,
+    bumpalo_ident: String,
+    bumpalo_owned_ident: String,
     builder_ident: String,
 }
 
 impl Message {
     fn try_new(m: &wrappers::Message) -> Result<Self> {
+        let mut bitfield_index = 0;
         let fields: Vec<Field> = m
             .fields()
             .into_iter()
-            .map(|f| Field::try_new(f))
+            .map(|f| Field::try_new(f, &mut bitfield_index))
             .try_collect()?;
-        let fields_len = fields.len();
         let oneofs = m
             .oneofs()
             .into_iter()
@@ -100,7 +103,6 @@ impl Message {
             .try_collect()?;
         Ok(Self {
             ident: m.rust_ident().to_string(),
-            proto_name: m.proto_name().to_string(),
             trait_ident: m.rust_trait_ident().to_string(),
             trait_delegate_macro_ident: format!("{}_delegate", m.rust_nested_module_ident()),
             submodule_ident: m.rust_nested_module_ident().to_string(),
@@ -109,10 +111,12 @@ impl Message {
                 enums: nested_enums,
             },
             fields,
-            fields_len,
             oneofs,
+            bitfield_len: bitfield_index,
             simple_ident: m.rust_impl_ident(""),
             single_field_ident: m.rust_impl_ident("SingleField"),
+            bumpalo_ident: m.rust_impl_ident("Bumpalo"),
+            bumpalo_owned_ident: m.rust_impl_ident("BumpaloOwned"),
             builder_ident: m.rust_impl_ident("Builder"),
         })
     }
@@ -168,96 +172,169 @@ impl EnumValue {
 struct Field {
     ident: String,
     ident_unesc: String,
-    proto_name: String,
+    ident_camel_unesc: String,
     number: i32,
     oneof_index: i32,
     is_message: bool,
     is_string: bool,
     is_bytes: bool,
     is_length_delimited: bool,
+    is_numerical: bool,
     is_explicit_oneof_field: bool,
     is_repeated: bool,
-    is_unlabeled: bool,
     has_default_value: bool,
     default_value: String,
+    has_optional_bit: bool,
+    bitfield_index: i32,
     trait_scalar_getter_type: String,
-    trait_maybe_field_message_trait_path: Option<String>,
+    trait_field_message_trait_path: String,
     trait_label_and_type_tags: String,
     oneof_enum_value_ident: String,
     simple_field_type: String,
+    simple_getter_type: String,
+    simple_getter_opt_type: String,
+    simple_getter_mut_type: String,
     simple_scalar_field_type: String,
-    simple_maybe_field_message_path: Option<String>,
-    simple_maybe_borrowed_field_type: Option<String>,
+    simple_field_message_path: String,
     simple_label_and_type_tags: String,
     single_field_type: String,
     single_numerical_rust_type: String,
+    bumpalo_field_type: String,
+    bumpalo_field_rep_item_type: String,
+    bumpalo_getter_type: String,
+    bumpalo_getter_opt_type: String,
+    bumpalo_getter_mut_type: String,
+    bumpalo_field_message_path: String,
+    bumpalo_label_and_type_tags: String,
 }
 
 impl Field {
-    fn try_new(f: &wrappers::Field) -> Result<Self> {
-        let trait_maybe_field_message_trait_path =
-            if let wrappers::FieldType::Message(m) = f.field_type()? {
-                Some(upgrade(&m)?.rust_trait_path())
-            } else {
-                None
-            };
-        let simple_maybe_field_message_path =
-            if let wrappers::FieldType::Message(m) = f.field_type()? {
-                Some(upgrade(&m)?.rust_impl_path("Simple"))
-            } else {
-                None
-            };
+    fn try_new(f: &wrappers::Field, bitfield_index: &mut i32) -> Result<Self> {
+        let maybe_message = if let wrappers::FieldType::Message(m) = f.field_type()? {
+            Some(upgrade(&m)?)
+        } else {
+            None
+        };
+        let is_message = matches!(f.field_type()?, wrappers::FieldType::Message(_));
+        let is_string = matches!(f.field_type()?, wrappers::FieldType::String);
+        let is_bytes = matches!(f.field_type()?, wrappers::FieldType::Bytes);
+        let is_repeated = matches!(f.field_label()?, wrappers::FieldLabel::Repeated);
+        let is_unlabeled = matches!(f.field_label()?, wrappers::FieldLabel::Unlabeled);
+        let is_length_delimited = is_message || is_string || is_bytes;
+        // i.e. any Option<> types except messages
+        let has_optional_bit = !is_repeated && !is_unlabeled && !is_message;
 
         Ok(Field {
-            ident: f.rust_ident().to_string(),
-            ident_unesc: f.rust_ident_unesc().to_string(),
-            proto_name: f.proto_name().to_string(),
+            ident: f.ident_lower_snake().to_string(),
+            ident_unesc: f.lower_snake_ident_unesc().to_string(),
+            ident_camel_unesc: f.ident_camel_unesc().to_string(),
             number: f.number(),
             oneof_index: f.oneof_index().unwrap_or(-1),
-            is_message: matches!(f.field_type()?, wrappers::FieldType::Message(_)),
-            is_string: matches!(f.field_type()?, wrappers::FieldType::String),
-            is_bytes: matches!(f.field_type()?, wrappers::FieldType::Bytes),
-            is_length_delimited: matches!(
-                f.field_type()?,
-                wrappers::FieldType::Bytes
-                    | wrappers::FieldType::String
-                    | wrappers::FieldType::Message(_)
-            ),
+            is_message,
+            is_string,
+            is_bytes,
+            is_length_delimited,
+            is_numerical: !is_length_delimited,
             is_explicit_oneof_field: f.oneof_index().is_some() && !f.is_optional3(),
-            is_repeated: matches!(f.field_label()?, wrappers::FieldLabel::Repeated),
-            is_unlabeled: matches!(f.field_label()?, wrappers::FieldLabel::Unlabeled),
+            is_repeated,
             has_default_value: f.default_value().is_some(),
             default_value: f
                 .default_value()
                 .map(|v| -> Result<_> { Ok(Self::convert_default_value(v, f.field_type()?)?) })
                 .transpose()?
-                .unwrap_or(Default::default()),
-            trait_scalar_getter_type: f.trait_scalar_getter_type()?,
-            trait_maybe_field_message_trait_path,
+                .unwrap_or("::std::default::Default::default()".to_string()),
+            has_optional_bit,
+            bitfield_index: {
+                if has_optional_bit {
+                    let i = *bitfield_index;
+                    *bitfield_index += 1;
+                    i
+                } else {
+                    -1
+                }
+            },
+            trait_scalar_getter_type: f.trait_scalar_getter_type()?.into(),
+            trait_field_message_trait_path: maybe_message
+                .as_ref()
+                .map(|m| m.rust_trait_path())
+                .unwrap_or_default(),
             trait_label_and_type_tags: f.rust_label_and_type_tags(|_| {
                 Ok(format!(
-                    "<Self as super::_puroro_traits::{trait_ident}>::Field{number}MessageType<'_>",
+                    "<Self as super::_puroro_traits::{trait_ident}>::{camel_ident}MessageType<'_>",
                     trait_ident = f.message()?.rust_trait_ident(),
-                    number = f.number(),
+                    camel_ident = f.ident_camel_unesc(),
                 ))
             })?,
-            oneof_enum_value_ident: f.rust_oneof_ident().to_string(),
-            simple_field_type: f.simple_field_type()?,
-            simple_scalar_field_type: f.simple_scalar_field_type()?,
-            simple_maybe_field_message_path,
-            simple_maybe_borrowed_field_type: f
-                .maybe_trait_scalar_getter_type_borrowed("Simple")?,
+            oneof_enum_value_ident: f.ident_camel().to_string(),
+            simple_field_type: f.simple_field_type()?.into(),
+            simple_getter_type: if f.is_repeated()? {
+                f.simple_getter_repeated_type("'_")?.into()
+            } else {
+                let bare_type = f.simple_getter_scalar_type("'_")?;
+                if f.is_message()? {
+                    format!("::std::option::Option<{}>", bare_type)
+                } else {
+                    bare_type.into()
+                }
+            },
+            simple_getter_opt_type: format!(
+                "::std::option::Option<{}>",
+                f.simple_getter_scalar_type("'_")?
+            ),
+            simple_getter_mut_type: f.simple_getter_mut_type("'_")?.into(),
+            simple_scalar_field_type: f.simple_scalar_field_type()?.into(),
+            simple_field_message_path: maybe_message
+                .as_ref()
+                .map(|m| m.rust_impl_path("Simple", &[]))
+                .unwrap_or_default(),
             simple_label_and_type_tags: f.rust_label_and_type_tags(|msg| {
                 Ok(
                     if matches!(f.field_label()?, wrappers::FieldLabel::Repeated) {
-                        msg.rust_impl_path("Simple")
+                        msg.rust_impl_path("Simple", &[])
                     } else {
-                        format!("::std::boxed::Box<{}>", msg.rust_impl_path("Simple"))
+                        format!("::std::boxed::Box<{}>", msg.rust_impl_path("Simple", &[]))
                     },
                 )
             })?,
             single_field_type: f.single_field_type()?,
-            single_numerical_rust_type: f.single_numerical_rust_type().unwrap_or("".to_string()),
+            single_numerical_rust_type: f.single_numerical_rust_type()?.into(),
+            bumpalo_field_type: f.bumpalo_field_type()?.into(),
+            bumpalo_field_rep_item_type: f.bumpalo_scalar_field_type()?.into(),
+            bumpalo_getter_type: if f.is_repeated()? {
+                f.bumpalo_getter_repeated_type("'this")?.to_string()
+            } else {
+                let bare_type = f.bumpalo_getter_scalar_type("'this")?;
+                if f.is_message()? {
+                    format!("::std::option::Option<{}>", bare_type)
+                } else {
+                    bare_type.to_string()
+                }
+            },
+            bumpalo_getter_opt_type: format!(
+                "::std::option::Option<{}>",
+                f.bumpalo_getter_scalar_type("'this")?
+            ),
+            bumpalo_getter_mut_type: if f.is_repeated()? {
+                f.bumpalo_getter_repeated_mut_type("'bump", "'this")?
+            } else {
+                f.bumpalo_getter_mut_type("'bump", "'this")?
+            },
+            bumpalo_field_message_path: maybe_message
+                .as_ref()
+                .map(|m| m.rust_impl_path("Bumpalo", &["'this"]))
+                .unwrap_or_default(),
+            bumpalo_label_and_type_tags: f.rust_label_and_type_tags(|msg| {
+                Ok(
+                    if matches!(f.field_label()?, wrappers::FieldLabel::Repeated) {
+                        msg.rust_impl_path("Bumpalo", &["'bump"])
+                    } else {
+                        format!(
+                            "::puroro::internal::NoAllocBumpBox<{}>",
+                            msg.rust_impl_path("Bumpalo", &["'bump"])
+                        )
+                    },
+                )
+            })?,
         })
     }
 
@@ -304,67 +381,8 @@ impl Field {
                 format!(r###""{}""###, input.escape_default().collect::<String>())
             }
             FieldType::Bytes => {
-                // protoc escapes 0x7F~0xFF character as octal escape "\234".
-                // Rust does not support that style so we need to re-encode it.
-                let mut decoded = Vec::new();
-                let mut bytes = input.bytes();
-                loop {
-                    if let Some(c) = bytes.next() {
-                        if c == b'\\' {
-                            if let Some(d) = bytes.next() {
-                                match d {
-                                    b'\\' | b'\"' | b'\'' => {
-                                        decoded.push(d);
-                                    }
-                                    b'r' => decoded.push(b'\r'),
-                                    b'n' => decoded.push(b'\n'),
-                                    b't' => decoded.push(b'\t'),
-                                    _ => {
-                                        let e_opt = bytes.next();
-                                        let f_opt = bytes.next();
-                                        match (d, e_opt, f_opt) {
-                                            (
-                                                (b'0'..=b'9'),
-                                                Some(e @ (b'0'..=b'9')),
-                                                Some(f @ (b'0'..=b'9')),
-                                            ) => {
-                                                let u8_value = u8::from_str_radix(
-                                                    &format!(
-                                                        "{}{}{}",
-                                                        d - b'0',
-                                                        e - b'0',
-                                                        f - b'0'
-                                                    ),
-                                                    8,
-                                                )
-                                                .map_err(|e| ErrorKind::ParseIntError {
-                                                    source: e,
-                                                })?;
-                                                decoded.push(u8_value);
-                                            }
-                                            _ => Err(ErrorKind::InvalidString {
-                                                string: input.to_string(),
-                                            })?,
-                                        }
-                                    }
-                                }
-                            } else {
-                                Err(ErrorKind::InvalidString {
-                                    string: input.to_string(),
-                                })?
-                            }
-                        } else {
-                            decoded.push(c);
-                        }
-                    } else {
-                        break;
-                    }
-                }
-                let reencoded = decoded
-                    .into_iter()
-                    .map(|b| format!(r"\x{:02x}", b))
-                    .collect::<String>();
-                format!(r#"b"{}""#, reencoded)
+                let rust_style_bytes = convert_octal_escape_to_rust_style_escape(input)?;
+                format!(r#"b"{}""#, rust_style_bytes)
             }
             FieldType::Enum2(e) | FieldType::Enum3(e) => {
                 format!(
@@ -385,6 +403,9 @@ impl Field {
 struct Oneof {
     index: i32,
     enum_ident: String,
+    simple_enum_ident: String,
+    bumpalo_enum_ident: String,
+    enum_maybe_gp_self: String,
     field_ident: String,
     fields: Vec<OneofField>,
     has_ld_field: bool,
@@ -394,27 +415,32 @@ struct Oneof {
 
 impl Oneof {
     fn try_new(o: &wrappers::Oneof) -> Result<Self> {
+        let has_ld_field = o.fields()?.into_iter().any(|f| {
+            matches!(
+                f.field_type(),
+                Ok(wrappers::FieldType::Bytes
+                    | wrappers::FieldType::String
+                    | wrappers::FieldType::Message(_))
+            )
+        });
+        let has_message_field = o
+            .fields()?
+            .into_iter()
+            .any(|f| matches!(f.field_type(), Ok(wrappers::FieldType::Message(_))));
         Ok(Oneof {
             index: o.index(),
             enum_ident: o.rust_enum_ident().to_string(),
+            simple_enum_ident: format!("{}Simple", o.rust_enum_ident().to_string()),
+            bumpalo_enum_ident: format!("{}Bumpalo", o.rust_enum_ident().to_string()),
+            enum_maybe_gp_self: if has_message_field { "<Self>" } else { "" }.to_string(),
             field_ident: o.rust_getter_ident().to_string(),
             fields: o
                 .fields()?
                 .into_iter()
                 .map(|f| OneofField::try_new(f))
                 .try_collect()?,
-            has_ld_field: o.fields()?.into_iter().any(|f| {
-                matches!(
-                    f.field_type(),
-                    Ok(wrappers::FieldType::Bytes
-                        | wrappers::FieldType::String
-                        | wrappers::FieldType::Message(_))
-                )
-            }),
-            has_message_field: o
-                .fields()?
-                .into_iter()
-                .any(|f| matches!(f.field_type(), Ok(wrappers::FieldType::Message(_)))),
+            has_ld_field,
+            has_message_field,
             owner_message_trait_path: o.message()?.rust_trait_path(),
         })
     }
@@ -425,37 +451,57 @@ struct OneofField {
     getter_ident: String,
     getter_ident_unesc: String,
     number: i32,
-    is_length_delimited: bool,
     is_message: bool,
+    is_string: bool,
+    is_bytes: bool,
+    is_length_delimited: bool,
+    is_numerical: bool,
     field_type: String,
+    simple_field_type: String,
+    simple_getter_opt_type: String,
+    simple_getter_mut_type: String,
+    bumpalo_field_type: String,
     trait_getter_type: String,
     simple_field_type_tag: String,
+    bumpalo_field_type_tag: String,
 }
 
 impl OneofField {
     fn try_new(f: &wrappers::Field) -> Result<Self> {
+        let is_message = matches!(f.field_type()?, wrappers::FieldType::Message(_));
+        let is_string = matches!(f.field_type()?, wrappers::FieldType::String);
+        let is_bytes = matches!(f.field_type()?, wrappers::FieldType::Bytes);
+        let is_length_delimited = is_message || is_string || is_bytes;
         Ok(Self {
-            ident: f.rust_oneof_ident().to_string(),
-            getter_ident: f.rust_ident().to_string(),
-            getter_ident_unesc: f.rust_ident_unesc().to_string(),
+            ident: f.ident_camel().to_string(),
+            getter_ident: f.ident_lower_snake().to_string(),
+            getter_ident_unesc: f.lower_snake_ident_unesc().to_string(),
             number: f.number(),
-            is_length_delimited: matches!(
-                f.field_type()?,
-                wrappers::FieldType::Bytes
-                    | wrappers::FieldType::String
-                    | wrappers::FieldType::Message(_)
+            is_message,
+            is_string,
+            is_bytes,
+            is_length_delimited,
+            is_numerical: !is_length_delimited,
+            field_type: f.trait_oneof_field_type("'msg", "T")?.into(),
+            simple_field_type: f.simple_field_type()?.into(),
+            simple_getter_opt_type: format!(
+                "::std::option::Option<{}>",
+                f.simple_getter_scalar_type("'_")?
             ),
-            is_message: matches!(f.field_type()?, wrappers::FieldType::Message(_)),
-            field_type: f.oneof_field_type()?,
-            trait_getter_type: f.trait_oneof_field_type("'this", "Self")?,
+            simple_getter_mut_type: f.simple_getter_mut_type("'_")?.into(),
+            bumpalo_field_type: f.bumpalo_oneof_field_type()?.into(),
+            trait_getter_type: f.trait_oneof_field_type("'this", "Self")?.into(),
             simple_field_type_tag: f.rust_type_tag(|msg| {
-                Ok(
-                    if matches!(f.field_label()?, wrappers::FieldLabel::Repeated) {
-                        msg.rust_impl_path("Simple")
-                    } else {
-                        format!("::std::boxed::Box<{}>", msg.rust_impl_path("Simple"))
-                    },
-                )
+                Ok(format!(
+                    "::std::boxed::Box<{}>",
+                    msg.rust_impl_path("Simple", &[])
+                ))
+            })?,
+            bumpalo_field_type_tag: f.rust_type_tag(|msg| {
+                Ok(format!(
+                    "::puroro::internal::NoAllocBumpBox<{}>",
+                    msg.rust_impl_path("Bumpalo", &["'bump"])
+                ))
             })?,
         })
     }
@@ -468,9 +514,15 @@ struct Structs<'a> {
 }
 
 #[derive(Template)]
-#[template(path = "traits.rs.txt")]
-struct Traits<'a> {
-    messages: &'a [Message],
+#[template(path = "trait.rs.txt")]
+struct Trait<'a> {
+    m: &'a Message,
+}
+
+#[derive(Template)]
+#[template(path = "private_oneof.rs.txt")]
+struct PrivateOneof<'a> {
+    oneof: &'a Oneof,
 }
 
 mod filters {
@@ -478,7 +530,10 @@ mod filters {
     pub(super) fn print_structs(messages: &[Message]) -> ::askama::Result<Structs> {
         Ok(Structs { messages })
     }
-    pub(super) fn print_traits(messages: &[Message]) -> ::askama::Result<Traits> {
-        Ok(Traits { messages })
+    pub(super) fn print_trait(message: &Message) -> ::askama::Result<Trait> {
+        Ok(Trait { m: message })
+    }
+    pub(super) fn print_private_oneof(oneof: &Oneof) -> ::askama::Result<PrivateOneof> {
+        Ok(PrivateOneof { oneof })
     }
 }
