@@ -21,6 +21,7 @@ use ::itertools::Itertools;
 use ::once_cell::unsync::{Lazy, OnceCell};
 use ::puroro_protobuf_compiled::google::protobuf;
 use ::std::borrow::Borrow;
+use ::std::borrow::Cow;
 use ::std::collections::{HashMap, VecDeque};
 use ::std::convert::TryInto;
 use ::std::hash::Hash;
@@ -33,6 +34,7 @@ use protobuf::{
     DescriptorProto, EnumDescriptorProto, FieldDescriptorProto, FileDescriptorProto,
     OneofDescriptorProto,
 };
+use puroro::tags::LengthDelimited;
 
 #[derive(Debug)]
 pub struct Context {
@@ -44,8 +46,7 @@ impl Context {
     pub fn try_from_proto(proto: CodeGeneratorRequest) -> Result<Rc<Context>> {
         let context = Rc::new_cyclic(|weak_context| Context {
             input_files: proto
-                .proto_file
-                .clone()
+                .proto_file()
                 .into_iter()
                 .map(|file| InputFile::try_from_proto(weak_context.clone(), file))
                 .collect::<Result<Vec<_>>>()
@@ -130,12 +131,10 @@ pub struct InputFile {
 }
 
 impl InputFile {
-    pub fn try_from_proto(context: Weak<Context>, proto: FileDescriptorProto) -> Result<Rc<Self>> {
+    pub fn try_from_proto(context: Weak<Context>, proto: &FileDescriptorProto) -> Result<Rc<Self>> {
         let package = Rc::new(
             proto
-                .package
-                .clone()
-                .unwrap_or_default()
+                .package()
                 .split('.')
                 .filter_map(|p| {
                     if p.is_empty() {
@@ -146,9 +145,9 @@ impl InputFile {
                 })
                 .collect_vec(),
         );
-        let proto_messages = proto.message_type.clone();
-        let proto_enums = proto.enum_type.clone();
-        let proto_syntax = proto.syntax.clone();
+        let proto_messages = proto.message_type();
+        let proto_enums = proto.enum_type();
+        let proto_syntax = proto.syntax();
         let mut file = Rc::new_cyclic(|file| Self {
             context: context,
             package: Clone::clone(&package),
@@ -176,10 +175,10 @@ impl InputFile {
                 })
                 .collect::<Result<Vec<_>>>()
                 .expect("I need try_new_cyclic..."),
-            syntax: match proto_syntax.as_deref() {
-                Some("proto2") | None => ProtoSyntax::Proto2,
-                Some("proto3") => ProtoSyntax::Proto3,
-                Some(syntax) => Err(ErrorKind::UnknownProtoSyntax {
+            syntax: match proto_syntax {
+                "proto2" | "" => ProtoSyntax::Proto2,
+                "proto3" => ProtoSyntax::Proto3,
+                syntax => Err(ErrorKind::UnknownProtoSyntax {
                     name: syntax.to_owned(),
                 })
                 .expect("I need try_new_cyclic..."),
@@ -226,26 +225,20 @@ pub struct Message {
 impl Message {
     pub fn try_from_proto(
         input_file: Weak<InputFile>,
-        proto: DescriptorProto,
+        proto: &DescriptorProto,
         package: Rc<Vec<String>>,
         outer_messages: Rc<Vec<String>>,
     ) -> Result<Rc<Self>> {
-        let proto_name = proto
-            .name
-            .clone()
-            .ok_or(ErrorKind::EmptyInputField {
-                name: "DescriptorProto.name".into(),
-            })?
-            .to_string();
+        let proto_name = proto.name().to_string();
         let new_outer_messages = Rc::new({
             let mut v = outer_messages.deref().clone();
             v.push(proto_name.clone());
             v
         });
-        let proto_field = proto.field.clone();
-        let proto_nested_type = proto.nested_type.clone();
-        let proto_enum_type = proto.enum_type.clone();
-        let proto_oneofs = proto.oneof_decl.clone();
+        let proto_field = proto.field();
+        let proto_nested_type = proto.nested_type();
+        let proto_enum_type = proto.enum_type();
+        let proto_oneofs = proto.oneof_decl();
         let message = Rc::new_cyclic(|message| Self {
             input_file: Clone::clone(&input_file),
             rust_ident: get_keyword_safe_ident(&to_camel_case(&proto_name)).to_string(),
@@ -316,18 +309,24 @@ impl Message {
             self.outer_messages.iter().map(|s| s.borrow()),
         )
     }
-    pub fn rust_impl_path(&self, impl_name: &str) -> String {
+    pub fn rust_impl_path(&self, impl_name: &str, gp: &[&str]) -> String {
         // "Simple" impls are separeted out to a special namespace.
         let module = if impl_name == "Simple" {
             "_puroro_simple_impl"
         } else {
             "_puroro_impls"
         };
+        let maybe_gp = if gp.is_empty() {
+            "".to_string()
+        } else {
+            format!("<{}>", gp.join(", "))
+        };
         format!(
-            "{path}::{module}::{ident}",
+            "{path}::{module}::{ident}{maybe_gp}",
             path = self.rust_module_path(),
             module = module,
             ident = self.rust_impl_ident(impl_name),
+            maybe_gp = maybe_gp,
         )
     }
 
@@ -398,18 +397,12 @@ pub struct Enum {
 impl Enum {
     pub fn try_from_proto(
         input_file: Weak<InputFile>,
-        proto: EnumDescriptorProto,
+        proto: &EnumDescriptorProto,
         package: Rc<Vec<String>>,
         outer_messages: Rc<Vec<String>>,
     ) -> Result<Rc<Self>> {
-        let proto_name = proto
-            .name
-            .clone()
-            .ok_or(ErrorKind::EmptyInputField {
-                name: "EnumDescriptorProto.name".into(),
-            })?
-            .to_string();
-        let proto_value = proto.value.clone();
+        let proto_name = proto.name().to_string();
+        let proto_value = proto.value();
         Ok(Rc::new(Self {
             input_file: input_file,
             rust_ident: get_keyword_safe_ident(&to_camel_case(&proto_name)).to_string(),
@@ -419,11 +412,8 @@ impl Enum {
             values: proto_value
                 .into_iter()
                 .map(|v| EnumValue {
-                    rust_ident: get_keyword_safe_ident(&to_camel_case(
-                        &v.name.clone().unwrap_or_default(),
-                    ))
-                    .to_string(),
-                    number: v.number.unwrap_or_default(),
+                    rust_ident: get_keyword_safe_ident(&to_camel_case(&v.name())).to_string(),
+                    number: v.number(),
                 })
                 .collect_vec(),
         }))
@@ -469,9 +459,10 @@ impl Enum {
 #[derive(Debug)]
 pub struct Field {
     message: Weak<Message>,
-    rust_ident: String,
-    rust_ident_unesc: String,
-    rust_oneof_ident: String,
+    ident_lower_snake: String,
+    ident_lower_snake_unesc: String,
+    ident_camel: String,
+    ident_camel_unesc: String,
     lazy_proto_type: OnceCell<FieldType>,
     proto_name: String,
     proto_type_name: String,
@@ -485,53 +476,48 @@ pub struct Field {
 }
 
 impl Field {
-    pub fn try_from_proto(message: Weak<Message>, proto: FieldDescriptorProto) -> Result<Self> {
-        let proto_name = proto
-            .name
-            .clone()
-            .ok_or(ErrorKind::EmptyInputField {
-                name: "FieldDescriptorProto.name".into(),
-            })?
-            .to_string();
-        let proto_type_name = proto.type_name.clone().unwrap_or_default().to_string();
-        let proto_type_enum = proto.r#type.clone().ok_or(ErrorKind::InternalError {
-            detail: "currently we are assuming the field type enum is always set.".to_string(),
-        })?;
-        let proto_label = proto.label.clone().ok_or(ErrorKind::InternalError {
-            detail: "currently we are assuming the field label enum is always set.".to_string(),
-        })?;
-        let proto_is_optional3 = proto.proto3_optional.unwrap_or_default();
-        let proto_number = proto.number.unwrap_or_default();
-        let proto_oneof_index = proto.oneof_index;
+    pub fn try_from_proto(message: Weak<Message>, proto: &FieldDescriptorProto) -> Result<Self> {
+        let proto_name = proto.name().to_string();
+        let proto_type_name = proto.type_name().to_string();
+        let proto_type_enum = proto.r#type();
+        let proto_label = proto.label();
+        let proto_is_optional3 = proto.proto3_optional();
+        let proto_number = proto.number();
+        let proto_oneof_index = proto.oneof_index_opt();
         Ok(Self {
             message: Clone::clone(&message),
-            rust_ident: get_keyword_safe_ident(&to_lower_snake_case(&proto_name)).to_string(),
-            rust_ident_unesc: to_lower_snake_case(&proto_name).to_string(),
-            rust_oneof_ident: get_keyword_safe_ident(&to_camel_case(&proto_name)).to_string(),
+            ident_lower_snake: get_keyword_safe_ident(&to_lower_snake_case(&proto_name))
+                .to_string(),
+            ident_lower_snake_unesc: to_lower_snake_case(&proto_name).to_string(),
+            ident_camel: get_keyword_safe_ident(&to_camel_case(&proto_name)).to_string(),
+            ident_camel_unesc: to_camel_case(&proto_name).to_string(),
             proto_name,
             proto_type_name,
             proto_type_enum,
             lazy_proto_type: OnceCell::new(),
             proto_label,
             proto_is_optional3,
-            proto_default_value: proto.default_value,
+            proto_default_value: proto.default_value_opt().map(|s| s.to_string()),
             lazy_label: OnceCell::new(),
             number: proto_number,
             proto_oneof_index,
         })
     }
 
-    pub fn rust_ident(&self) -> &str {
-        &self.rust_ident
-    }
-    pub fn rust_ident_unesc(&self) -> &str {
-        &self.rust_ident_unesc
-    }
     pub fn proto_name(&self) -> &str {
         &self.proto_name
     }
-    pub fn rust_oneof_ident(&self) -> &str {
-        &self.rust_oneof_ident
+    pub fn ident_lower_snake(&self) -> &str {
+        &self.ident_lower_snake
+    }
+    pub fn lower_snake_ident_unesc(&self) -> &str {
+        &self.ident_lower_snake_unesc
+    }
+    pub fn ident_camel(&self) -> &str {
+        &self.ident_camel
+    }
+    pub fn ident_camel_unesc(&self) -> &str {
+        &self.ident_camel_unesc
     }
     pub fn message(&self) -> Result<Rc<Message>> {
         Ok(upgrade(&self.message)?)
@@ -569,6 +555,18 @@ impl Field {
             })
             .map(|l| l.clone())
     }
+    pub fn is_repeated(&self) -> Result<bool> {
+        Ok(matches!(self.field_label()?, FieldLabel::Repeated))
+    }
+    pub fn is_message(&self) -> Result<bool> {
+        Ok(matches!(self.field_type()?, FieldType::Message(_)))
+    }
+    pub fn is_length_delimited(&self) -> Result<bool> {
+        Ok(matches!(
+            self.field_type()?,
+            FieldType::Message(_) | FieldType::String | FieldType::Bytes
+        ))
+    }
     pub fn is_optional3(&self) -> bool {
         self.proto_is_optional3
     }
@@ -617,44 +615,47 @@ impl Field {
         ))
     }
 
-    pub fn trait_scalar_getter_type(&self) -> Result<String> {
-        Ok(match self.field_type()? {
-            FieldType::Group => Err(ErrorKind::GroupNotSupported)?,
-            FieldType::Enum2(e) | FieldType::Enum3(e) => upgrade(&e)?.rust_path(),
-            FieldType::String => "&'this str".to_string(),
-            FieldType::Bytes => "&'this [u8]".to_string(),
-            FieldType::Message(_) => format!(
-                "Self::Field{number}MessageType<'this>",
-                number = self.number()
-            ),
-            t => t.numerical_rust_type()?.to_string(),
+    pub fn trait_scalar_getter_type(&self) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => "&'this str".into(),
+            LengthDelimited(Bytes) => "&'this [u8]".into(),
+            LengthDelimited(Message(_)) => format!(
+                "Self::{ident_camel}MessageType<'this>",
+                ident_camel = self.ident_camel_unesc
+            )
+            .into(),
+            Trivial(field_type) => field_type.rust_type_name()?,
         })
     }
-    pub fn trait_oneof_field_type(&self, lt: &str, trait_impl: &str) -> Result<String> {
-        Ok(match self.field_type()? {
-            FieldType::Group => Err(ErrorKind::GroupNotSupported)?,
-            FieldType::Enum2(e) | FieldType::Enum3(e) => upgrade(&e)?.rust_path(),
-            FieldType::String => format!("&{lt} str", lt = lt),
-            FieldType::Bytes => format!("&{lt} [u8]", lt = lt),
-            FieldType::Message(_) => format!(
-                "<{trait_impl} as {trait_path}>::Field{number}MessageType<{lt}>",
+    pub fn trait_oneof_field_type(&self, lt: &str, trait_impl: &str) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => format!("&{lt} str", lt = lt).into(),
+            LengthDelimited(Bytes) => format!("&{lt} [u8]", lt = lt).into(),
+            LengthDelimited(Message(_)) => format!(
+                "<{trait_impl} as {trait_path}>::{ident_camel}MessageType<{lt}>",
                 lt = lt,
                 trait_impl = trait_impl,
                 trait_path = self.message()?.rust_trait_path(),
-                number = self.number(),
-            ),
-            t => t.numerical_rust_type()?.to_string(),
+                ident_camel = self.ident_camel_unesc,
+            )
+            .into(),
+            Trivial(field_type) => field_type.rust_type_name()?,
         })
     }
 
     pub fn maybe_trait_scalar_getter_type_borrowed(
         &self,
         impl_name: &str,
+        gp: &[&str],
     ) -> Result<Option<String>> {
         Ok(match self.field_type() {
             Ok(FieldType::String) => Some("str".to_string()),
             Ok(FieldType::Bytes) => Some("[u8]".to_string()),
-            Ok(FieldType::Message(m)) => Some(upgrade(&m)?.rust_impl_path(impl_name)),
+            Ok(FieldType::Message(m)) => Some(upgrade(&m)?.rust_impl_path(impl_name, gp)),
             _ => None,
         })
     }
@@ -683,78 +684,241 @@ impl Field {
         Ok(matches!(self.field_label(), Ok(FieldLabel::Repeated)))
     }
 
-    pub fn oneof_field_type(&self) -> Result<String> {
-        Ok(match self.field_type()? {
-            FieldType::Group => Err(ErrorKind::GroupNotSupported)?,
-            FieldType::Enum2(e) | FieldType::Enum3(e) => upgrade(&e)?.rust_path(),
-            FieldType::String => "<IsOwned as ::puroro::internal::bool::BoolType>\
-                ::Choose<::std::string::String, &'msg str>"
-                .to_string(),
-            FieldType::Bytes => "<IsOwned as ::puroro::internal::bool::BoolType>\
-                ::Choose<::std::vec::Vec<u8>, &'msg [u8]>"
-                .to_string(),
-            FieldType::Message(m) => {
-                let simple_message_type = upgrade(&m)?.rust_impl_path("Simple");
-                let trait_getter_type = format!(
-                    "<T as {trait_path}>::Field{number}MessageType<'msg>",
-                    trait_path = self.message()?.rust_trait_path(),
-                    number = self.number(),
-                );
+    pub fn bumpalo_oneof_field_type(&self) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        let bare_type = match self.field_type()?.categories()? {
+            LengthDelimited(String) => "::puroro::internal::NoAllocBumpString".into(),
+            LengthDelimited(Bytes) => "::puroro::internal::NoAllocBumpVec<u8>".into(),
+            LengthDelimited(Message(m)) => {
+                let bumpalo_message_type = upgrade(&m)?.rust_impl_path("Bumpalo", &["'bump"]);
                 format!(
-                    "<IsOwned as ::puroro::internal::bool::BoolType>\
-                    ::Choose<::std::boxed::Box<{message_type}>, {trait_getter_type}>",
-                    message_type = simple_message_type,
-                    trait_getter_type = trait_getter_type,
+                    "::puroro::internal::NoAllocBumpBox<{message_type}>",
+                    message_type = bumpalo_message_type,
                 )
+                .into()
             }
-            t => t.numerical_rust_type()?.to_string(),
-        })
+            Trivial(field_type) => field_type.rust_type_name()?,
+        };
+        Ok(format!("::puroro::internal::Bare<{}>", bare_type).into())
     }
 
-    pub fn simple_field_type(&self) -> Result<String> {
+    pub fn simple_field_type(&self) -> Result<Cow<'static, str>> {
         let scalar_type = self.simple_scalar_field_type()?;
         Ok(match self.field_label()? {
-            FieldLabel::OneofField => scalar_type,
-            FieldLabel::Required | FieldLabel::Optional => {
-                format!("::std::option::Option<{}>", scalar_type)
-            }
-            FieldLabel::Unlabeled => {
+            FieldLabel::Repeated => format!("::std::vec::Vec<{}>", scalar_type).into(),
+            FieldLabel::OneofField => format!("::puroro::internal::Bare<{}>", scalar_type).into(),
+            _ => {
                 if matches!(self.field_type(), Ok(FieldType::Message(_))) {
-                    format!("::std::option::Option<{}>", scalar_type)
+                    format!("::std::option::Option<{}>", scalar_type).into()
                 } else {
-                    scalar_type
+                    format!("::puroro::internal::Bare<{}>", scalar_type).into()
                 }
             }
-            FieldLabel::Repeated => format!("::std::vec::Vec<{}>", scalar_type),
         })
     }
 
-    pub fn simple_oneof_field_type(&self) -> Result<String> {
-        let scalar_type = self.simple_scalar_field_type()?;
-        Ok(scalar_type)
-    }
-
-    pub fn simple_scalar_field_type(&self) -> Result<String> {
-        Ok(match self.field_type()? {
-            FieldType::Group => Err(ErrorKind::GroupNotSupported)?,
-            FieldType::String => "::std::string::String".to_string(),
-            FieldType::Bytes => "::std::vec::Vec<u8>".to_string(),
-            FieldType::Enum2(e) => upgrade(&e)?.rust_path(),
-            FieldType::Enum3(e) => upgrade(&e)?.rust_path(),
-            FieldType::Message(m) => {
-                let bare_msg = upgrade(&m)?.rust_impl_path("Simple");
+    pub fn simple_scalar_field_type(&self) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => "::std::string::String".into(),
+            LengthDelimited(Bytes) => "::std::vec::Vec<u8>".into(),
+            LengthDelimited(Message(m)) => {
+                let bare_msg = upgrade(&m)?.rust_impl_path("Simple", &[]);
                 if matches!(self.field_label(), Ok(FieldLabel::Repeated)) {
-                    bare_msg
+                    bare_msg.into()
                 } else {
-                    format!("::std::boxed::Box<{}>", bare_msg)
+                    format!("::std::boxed::Box<{}>", bare_msg).into()
                 }
             }
-            t => t.numerical_rust_type()?.to_string(),
+            Trivial(field_type) => field_type.rust_type_name()?,
+        })
+    }
+
+    pub fn simple_getter_scalar_type(&self, lt: &str) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => format!("&{} str", lt).into(),
+            LengthDelimited(Bytes) => format!("&{} [u8]", lt).into(),
+            LengthDelimited(Message(m)) => {
+                let msg_type = upgrade(&m)?.rust_impl_path("Simple", &[]);
+                format!("&{lt} {msg}", lt = lt, msg = msg_type).into()
+            }
+            Trivial(field_type) => field_type.rust_type_name()?,
+        })
+    }
+
+    pub fn simple_getter_repeated_type(&self, lt: &str) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => {
+                format!(
+                    "&{lt}[impl ::std::ops::Deref<Target=str> + ::std::fmt::Debug]",
+                    lt = lt
+                )
+            }
+            LengthDelimited(Bytes) => {
+                format!(
+                    "&{lt}[impl ::std::ops::Deref<Target=[u8]> + ::std::fmt::Debug]",
+                    lt = lt
+                )
+            }
+            LengthDelimited(Message(m)) => {
+                let msg_type = upgrade(&m)?.rust_impl_path("Simple", &[]);
+                format!("&{lt}[{msg}]", lt = lt, msg = msg_type)
+            }
+            Trivial(field_type) => {
+                format!("&{lt}[{ty}]", lt = lt, ty = field_type.rust_type_name()?)
+            }
+        }
+        .into())
+    }
+
+    pub fn simple_getter_mut_type(&self, lt: &str) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok({
+            let scalar_type = match self.field_type()?.categories()? {
+                LengthDelimited(String) => "::std::string::String".into(),
+                LengthDelimited(Bytes) => "::std::vec::Vec<u8>".into(),
+                LengthDelimited(Message(m)) => upgrade(&m)?.rust_impl_path("Simple", &[]).into(),
+                Trivial(field_type) => field_type.rust_type_name()?,
+            };
+            if self.is_repeated()? {
+                format!("&{lt} mut ::std::vec::Vec<{ty}>", lt = lt, ty = scalar_type)
+            } else {
+                format!("&{lt} mut {ty}", lt = lt, ty = scalar_type)
+            }
+            .into()
+        })
+    }
+
+    pub fn bumpalo_field_type(&self) -> Result<Cow<'static, str>> {
+        let scalar_type = self.bumpalo_scalar_field_type()?;
+        if self.is_repeated()? {
+            Ok(format!("::puroro::internal::NoAllocBumpVec<{}>", scalar_type).into())
+        } else if self.is_message()? {
+            Ok(format!(
+                "::std::option::Option<::puroro::internal::NoAllocBumpBox<{}>>",
+                scalar_type
+            )
+            .into())
+        } else {
+            Ok(format!("::puroro::internal::Bare<{}>", scalar_type).into())
+        }
+    }
+
+    pub fn bumpalo_scalar_field_type(&self) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => "::puroro::internal::NoAllocBumpString".into(),
+            LengthDelimited(Bytes) => "::puroro::internal::NoAllocBumpVec<u8>".into(),
+            LengthDelimited(Message(m)) => {
+                upgrade(&m)?.rust_impl_path("Bumpalo", &["'bump"]).into()
+            }
+            Trivial(field_type) => field_type.rust_type_name()?,
+        })
+    }
+
+    pub fn bumpalo_getter_scalar_type(&self, lt: &str) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => format!("&{} str", lt).into(),
+            LengthDelimited(Bytes) => format!("&{} [u8]", lt).into(),
+            LengthDelimited(Message(m)) => {
+                let msg_type = upgrade(&m)?.rust_impl_path("Bumpalo", &[lt]);
+                format!("&{lt} {msg}", lt = lt, msg = msg_type).into()
+            }
+            Trivial(field_type) => field_type.rust_type_name()?,
+        })
+    }
+
+    pub fn bumpalo_getter_repeated_type(&self, lt: &str) -> Result<Cow<'static, str>> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => {
+                format!("&{lt}[impl ::std::ops::Deref<Target=str>]", lt = lt)
+            }
+            LengthDelimited(Bytes) => {
+                format!("&{lt}[impl ::std::ops::Deref<Target=[u8]>]", lt = lt)
+            }
+            LengthDelimited(Message(m)) => {
+                let msg_type = upgrade(&m)?.rust_impl_path("Bumpalo", &[lt]);
+                format!("&{lt}[{msg}]", lt = lt, msg = msg_type)
+            }
+            Trivial(field_type) => {
+                format!("&{lt}[{ty}]", lt = lt, ty = field_type.rust_type_name()?)
+            }
+        }
+        .into())
+    }
+
+    pub fn bumpalo_getter_mut_type(&self, bump_lt: &str, this_lt: &str) -> Result<String> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            // ↓ Maybe need to double check the lt params
+            LengthDelimited(String) => format!(
+                "impl {this_lt} + ::std::ops::DerefMut<Target=::puroro::bumpalo::collections::String<{bump_lt}>>",
+                bump_lt = bump_lt,
+                this_lt = this_lt,
+            ),
+            LengthDelimited(Bytes) => format!(
+                "impl {this_lt} + ::std::ops::DerefMut<Target=::puroro::bumpalo::collections::Vec<{bump_lt}, u8>>",
+                bump_lt = bump_lt,
+                this_lt = this_lt,
+            ),
+            LengthDelimited(Message(m)) => {
+                let msg_type = upgrade(&m)?.rust_impl_path("Bumpalo", &[bump_lt]);
+                format!("&{lt} mut {msg}", lt = this_lt, msg = msg_type)
+            }
+            Trivial(field_type) => format!(
+                "&{lt} mut {ty}",
+                lt = this_lt,
+                ty = field_type.rust_type_name()?
+            ),
+        })
+    }
+
+    pub fn bumpalo_getter_repeated_mut_type(&self, bump_lt: &str, this_lt: &str) -> Result<String> {
+        use FieldTypeCategories::*;
+        use LdFieldType::*;
+        Ok(match self.field_type()?.categories()? {
+            LengthDelimited(String) => format!(
+                "::puroro::internal::AddBumpVecView<{bump_lt}, {this_lt}, ::puroro::internal::NoAllocBumpString>",
+                bump_lt = bump_lt,
+                this_lt = this_lt
+            ),
+            LengthDelimited(Bytes) => format!(
+                "::puroro::internal::AddBumpVecView<{bump_lt}, {this_lt}, ::puroro::internal::NoAllocBumpVec<u8>>",
+                bump_lt = bump_lt,
+                this_lt = this_lt
+            ),
+            LengthDelimited(Message(m)) => format!(
+                "::puroro::internal::RefMutBumpVec<{bump_lt}, {this_lt}, {ty}>",
+                bump_lt = bump_lt,
+                this_lt = this_lt,
+                ty = upgrade(&m)?.rust_impl_path("Bumpalo", &[bump_lt])
+            ),
+            Trivial(field_type) => {
+                format!(
+                    "::puroro::internal::RefMutBumpVec<{bump_lt}, {this_lt}, {ty}>",
+                    bump_lt = bump_lt,
+                    this_lt = this_lt,
+                    ty = field_type.rust_type_name()?
+                )
+            }
         })
     }
 
     pub fn single_field_type(&self) -> Result<String> {
-        Ok(if matches!(self.field_label()?, FieldLabel::Repeated) {
+        Ok(if self.is_repeated()? {
             "RepeatedType"
         } else {
             "ScalarType"
@@ -762,11 +926,10 @@ impl Field {
         .to_string())
     }
 
-    pub fn single_numerical_rust_type(&self) -> Result<String> {
-        Ok(match self.field_type()? {
-            FieldType::Group => Err(ErrorKind::GroupNotSupported)?,
-            FieldType::Enum2(e) | FieldType::Enum3(e) => upgrade(&e)?.rust_path(),
-            t => t.numerical_rust_type()?.to_string(),
+    pub fn single_numerical_rust_type(&self) -> Result<Cow<'static, str>> {
+        Ok(match self.field_type()?.categories()? {
+            FieldTypeCategories::Trivial(field_type) => field_type.rust_type_name()?,
+            _ => "".into(),
         })
     }
 }
@@ -784,23 +947,18 @@ pub struct Oneof {
 impl Oneof {
     pub fn try_from_proto(
         message: Weak<Message>,
-        proto: OneofDescriptorProto,
+        proto: &OneofDescriptorProto,
         index: i32,
     ) -> Result<Rc<Self>> {
-        if let Some(name) = proto.name.as_ref() {
-            Ok(Rc::new(Self {
-                message,
-                index: index,
-                rust_enum_ident: get_keyword_safe_ident(&to_camel_case(name)).to_string(),
-                rust_getter_ident: get_keyword_safe_ident(&to_lower_snake_case(name)).to_string(),
-                lazy_fields: OnceCell::new(),
-                lazy_is_synthetic: OnceCell::new(),
-            }))
-        } else {
-            Err(ErrorKind::EmptyInputField {
-                name: "OneofDescriptorProto.name".to_string(),
-            })?
-        }
+        Ok(Rc::new(Self {
+            message,
+            index: index,
+            rust_enum_ident: get_keyword_safe_ident(&to_camel_case(proto.name())).to_string(),
+            rust_getter_ident: get_keyword_safe_ident(&to_lower_snake_case(proto.name()))
+                .to_string(),
+            lazy_fields: OnceCell::new(),
+            lazy_is_synthetic: OnceCell::new(),
+        }))
     }
     pub fn rust_enum_ident(&self) -> &str {
         &self.rust_enum_ident
@@ -1022,24 +1180,28 @@ impl FieldType {
         })
     }
 
-    pub fn numerical_rust_type(&self) -> Result<&'static str> {
-        Ok(match *self {
-            FieldType::Double => "f64",
-            FieldType::Float => "f32",
-            FieldType::Int32 => "i32",
-            FieldType::Int64 => "i64",
-            FieldType::UInt32 => "u32",
-            FieldType::UInt64 => "u64",
-            FieldType::SInt32 => "i32",
-            FieldType::SInt64 => "i64",
-            FieldType::Fixed32 => "u32",
-            FieldType::Fixed64 => "u64",
-            FieldType::SFixed32 => "i32",
-            FieldType::SFixed64 => "i64",
-            FieldType::Bool => "bool",
-            _ => Err(ErrorKind::InternalError {
-                detail: "numerical_rust_type() is invoked for non-numerical type".to_string(),
-            })?,
+    pub fn categories(self) -> Result<FieldTypeCategories> {
+        use FieldTypeCategories::*;
+        Ok(match self {
+            FieldType::Double => Trivial(TrivialFieldType::Double),
+            FieldType::Float => Trivial(TrivialFieldType::Float),
+            FieldType::Int32 => Trivial(TrivialFieldType::Int32),
+            FieldType::Int64 => Trivial(TrivialFieldType::Int64),
+            FieldType::UInt32 => Trivial(TrivialFieldType::UInt32),
+            FieldType::UInt64 => Trivial(TrivialFieldType::UInt64),
+            FieldType::SInt32 => Trivial(TrivialFieldType::SInt32),
+            FieldType::SInt64 => Trivial(TrivialFieldType::SInt64),
+            FieldType::Fixed32 => Trivial(TrivialFieldType::Fixed32),
+            FieldType::Fixed64 => Trivial(TrivialFieldType::Fixed64),
+            FieldType::SFixed32 => Trivial(TrivialFieldType::SFixed32),
+            FieldType::SFixed64 => Trivial(TrivialFieldType::SFixed64),
+            FieldType::Bool => Trivial(TrivialFieldType::Bool),
+            FieldType::Group => Err(ErrorKind::GroupNotSupported)?,
+            FieldType::String => LengthDelimited(LdFieldType::String),
+            FieldType::Bytes => LengthDelimited(LdFieldType::Bytes),
+            FieldType::Enum2(e) => Trivial(TrivialFieldType::Enum2(e)),
+            FieldType::Enum3(e) => Trivial(TrivialFieldType::Enum3(e)),
+            FieldType::Message(m) => LengthDelimited(LdFieldType::Message(m)),
         })
     }
 
@@ -1068,6 +1230,58 @@ impl FieldType {
     }
 }
 
+pub enum LdFieldType {
+    String,
+    Bytes,
+    Message(Weak<Message>),
+}
+
+pub enum TrivialFieldType {
+    Double,
+    Float,
+    Int32,
+    Int64,
+    UInt32,
+    UInt64,
+    SInt32,
+    SInt64,
+    Fixed32,
+    Fixed64,
+    SFixed32,
+    SFixed64,
+    Bool,
+    Enum2(Weak<Enum>),
+    Enum3(Weak<Enum>),
+}
+
+impl TrivialFieldType {
+    pub fn rust_type_name(&self) -> Result<Cow<'static, str>> {
+        Ok(match self {
+            TrivialFieldType::Double => "f64".into(),
+            TrivialFieldType::Float => "f32".into(),
+            TrivialFieldType::Int32 => "i32".into(),
+            TrivialFieldType::Int64 => "i64".into(),
+            TrivialFieldType::UInt32 => "u32".into(),
+            TrivialFieldType::UInt64 => "u64".into(),
+            TrivialFieldType::SInt32 => "i32".into(),
+            TrivialFieldType::SInt64 => "i64".into(),
+            TrivialFieldType::Fixed32 => "u32".into(),
+            TrivialFieldType::Fixed64 => "u64".into(),
+            TrivialFieldType::SFixed32 => "i32".into(),
+            TrivialFieldType::SFixed64 => "i64".into(),
+            TrivialFieldType::Bool => "bool".into(),
+            TrivialFieldType::Enum2(e) | TrivialFieldType::Enum3(e) => {
+                upgrade(e)?.rust_path().into()
+            }
+        })
+    }
+}
+
+pub enum FieldTypeCategories {
+    LengthDelimited(LdFieldType),
+    Trivial(TrivialFieldType),
+}
+
 #[derive(Debug, Clone)]
 pub enum FieldLabel {
     Required,
